<?php

/**
 * @file
 * Holds the contents of a preprocess function moved into its own file
 * to ease memory requirements and having too much code in one file.
 */

/*
 * Simplistic HTML attribute extraction.
 */
function _itweak_links_extract_attrs($str, $class_array = FALSE) {
// This function will choke on attr without ="...", i.e 'attr1 attr2="val"'.
  $attrs = array();
  $pattern = '/(\w+)\s*=\s*("[^"]*"|\'[^\']*\'|[^"\'\s>]*)/';
  preg_match_all($pattern, $str, $matches, PREG_SET_ORDER);
  foreach ($matches as $match) {
    if (($match[2][0] == '"' || $match[2][0] == "'") && $match[2][0] == $match[2][strlen($match[2])-1]) {
      $match[2] = substr($match[2], 1, -1);
    }
    $name = strtolower($match[1]);
    $value = html_entity_decode($match[2]);
    switch ($name) {
    case 'class':
      $attrs[$name] = $class_array ? preg_split('/\s+/', trim($value)) : $value;
      break;
    default:
      $attrs[$name] = $value;
    }
  }
//watchdog('_extract_attrs','str='.iprintr($str).' out='.iprintr($attrs));
  return $attrs;
}

/*
 * Simplistic HTML attribute merge into a tag.
 */
function _itweak_links_insert_attributes($html, $attributes, $tag = 'a') {
  // Make sure class attribute is an array for array_merge() below.
  if (!empty($attributes['class']) && !is_array($attributes['class'])) {
    $attributes['class'] = preg_split('/\s+/', $attributes['class']);
  }

  // Find where the insides of each $tag are in $html
  // Limitations: Will break if there is any unencoded '>' inside attribute value strings.
  $pattern = '/<' . $tag . '\s+([^>]*)>/mi';
  preg_match_all($pattern, $html, $matches, PREG_SET_ORDER | PREG_OFFSET_CAPTURE);

  // Prepare list of replacement positions, merge attributes
  $elems = array();
  foreach ($matches as $match) {
    $orig_attrs = _itweak_links_extract_attrs($match[1][0], FALSE);
    $attrs = array_merge($orig_attrs, $attributes);
    if (!empty($orig_attrs['class']) && !empty($attributes['class'])) {
       $orig_attrs['class'] = preg_split('/\s+/', trim($orig_attrs['class']));
       $attrs['class'] = implode(' ', array_unique(array_merge($orig_attrs['class'], $attributes['class'])));
    }
    $elems[] = array(drupal_attributes($attrs), $match[1][1], strlen($match[1][0]));
  }

  // Apply substr_replace() in reverse order so each replacement will not invalidate lower offsets
  $elems = array_reverse($elems);
  $html_out = $html;
  foreach ($elems as $elem) {
    $html_out = substr_replace($html_out, $elem[0], $elem[1], $elem[2]);
  }

//watchdog('insert attributes','in='.iprintr($html).' out='.iprintr($html_out).' matches='.iprintr($matches).' elems='.iprintr($elems));
  return $html_out;
}

/**
 * Processing of node in pre-render stage.
 * Move some links into the node.
 */
function _itweak_links_node_pre_render($content) {
  $node = &$content['itl_prefix']['#node'];
  $content['itl_prefix']['#value'] = '';
  $content['itl_suffix']['#value'] = '';

  if (is_array($node->links)) {
    $links = &$node->links;
    foreach ($links as $link_key => $link) {
      $class = $link_key;
      if (isset($link['itl_prerender'])) {
        $options = $link['itl_prerender'];
        if ($options['mode'] == 'before_node' || $options['mode'] == 'after_node' || $options['mode'] == 'inline_node') {

          $html = '';
          if (isset($link['href'])) {
            // Pass in $link as $options, they share the same keys.
            $html .= l($link['title'], $link['href'], $link);
          }
          else if (!empty($link['title'])) {
            // Some links are actually not links, but we wrap these in <span> for adding title and class attributes
            if (empty($link['html'])) {
              $link['title'] = check_plain($link['title']);
            }
            $span_attributes = '';
            if (isset($link['attributes'])) {
              $span_attributes = drupal_attributes($link['attributes']);
            }
            $html .= '<span' . $span_attributes . '>' . $link['title'] . '</span>';
          }
          $html .= "\n";
          if ($options['mode'] == 'before_node') {
            $content['itl_prefix']['#value'] .= $html;
          }
          elseif ($options['mode'] == 'after_node') {
            $content['itl_suffix']['#value'] .= $html;
          }
          else { // $options['mode'] == 'inline_node'
            $elements_array = (!empty($options['inline_node_elements'])) ? $options['inline_node_elements'] : array('p');
            $elements = '(?:' . implode('|', $elements_array) . ')';
            // Get last position of the last closing marker in teaser
            if (preg_match('!</?'. $elements .'[^>]*>$!i', $content['body']['#value'], $match, PREG_OFFSET_CAPTURE)) {
              // Recalculate the position in body. We do this because there may be extra CCK fields appended to the teaser.
              // (This bit is copied from ed_readmore.module, and node->teaser swapped with content[body] as this seems to be the right sequence of rendering in drupal core.)
              $insert_point = strpos($content['body']['#value'], $node->teaser) + $match[0][1];
              // Insert the link
              $content['body']['#value'] = substr_replace($content['body']['#value'], ' ' . $html, $insert_point, 0);
            }
            else {
              // Not found, fallback to 'after_node'
              $content['itl_suffix']['#value'] .= $html;
            }
          }
          unset($links[$link_key]);
        }
      }
    }
  }
  if (empty($content['itl_prefix']['#value'])) {
    unset($content['itl_prefix']);
  }
  if (empty($content['itl_suffix']['#value'])) {
    unset($content['itl_suffix']);
  }
  return $content;
}

/**
 * Processing of node in nodeapi view stage.
 */
function _itweak_links_node_view_prepare(&$node, $op, $teaser, $page) {
  // Insert two elements to locate points before and after the node body
  $node->content['itl_prefix'] = array(
    '#weight' => -1000000,
    '#value' => '',
  );
  $node->content['itl_suffix'] = array(
    '#weight' => +1000000,
    '#value' => '',
  );

  // Stash the pointer to the node inside content for _itweak_links_node_pre_render()
  $node->content['itl_prefix']['#node'] = &$node;

  // Attach our pre-render callback
  if (!isset($node->content['#pre_render'])) {
    $node->content['#pre_render'] = array();
  }
  $node->content['#pre_render'][] = '_itweak_links_node_pre_render';
}

/**
 * Helper function to compare arrays.
 */
function _itweak_links_link_cmp($a, $b) {
  if ($a['weight'] == $b['weight']) {
    return 0;
  }
  return ($a['weight'] < $b['weight']) ? -1 : 1;
}

/**
 * This is the place for customization.
 * Currently hard-coded, but can be developed into admin input feature.
 */
function _itweak_links_preprocess_links_settings() {
  static $return = '';
  if (!$return) {
    $settings = array(
    // This hash array lists all the links we want to modify in their desired order
    // The array is keyed by record id.
    // The array elements are hash arrays of options.
    // Options array have mandatory 'pattern' and 'mode' and a number of optional elements.
    // mandatory 'pattern' is a regular expression (for convenience, % is substituted with .* and ! is substituted with \! automatically)
    // optional 'key' changes what to match the pattern to, e.g. 'key'=>'href' will check the link's href instead of its name
    // optional 'copy' => TRUE will make this entry if matched to be the last one (stop matching)
    // optional 'type' and 'node_type' restrict matching based on $type and $node_type given by the caller
    // mandatory 'mode' is one of:
    // - 'link' (default) - leave as link,
    // - 'none'           - remove completely,
    // - 'cssbutton'      - format as a CSS button / or image replacement (with <span> tag insertion),
    // - 'dropdown'       - place into drop-down menu (optional 'dropdown_name' gives menu name)
    // - 'before_node'    - move the link on a separate line before the body
    // - 'after_node'     - move the link to a new line after the body
    // - 'inline_node'    - move the link after the last word of the body, before last closing tag
    // optional 'weight' moves the weight of that item out of default
    //  order - most usefull to put the link ahead of all other links (negative weight).
    // If 'weight' is not given, all listed/matched links are moved to the end of the list.
    // Note that CSS can float:right some links, so their order will appear reverse to the one given here.
    // optional 'href' element replaces original href of the link.
    // optional 'class' element defines CSS class for the link, appended by default and replaces the original if 'class_replace' => TRUE.
    // optional 'class_replace' if TRUE makes 'class' replace the original link class.
    // For 'mode' => 'dropdown', optional 'dropdown_class' element defines CSS class for the drop-down div.
    //   If not given, 'dropdown_class' defaults to 'itweak-links-dropdown'. Three additional derived class names are used: -title, -title-wrap, -menu.
    // optional 'text' will replace link text ('text_html' = TRUE will set $link->html to TRUE, otherwise will check_plain).
    // optional 'text_prefix' and/or 'text_suffix' element defines HTML code to add before/after link text, $link->html is set to TRUE then.
    // NOT SUPPPORTED on D6: optional 'prefix' and/or 'suffix' element defines HTML code to add before/after the link.
    // The following options translate to link attributes (I decided not to make a sub-array to simplify settings database):
    // optional 'title' will replace link title (title attr inside <a> tag)
    // optional 'nofollow' = TRUE will add rel="nofollow" attribute. To simplify data entry here, 'nofollow' = FALSE will override default 'nofollow' = TRUE that is added in the preprocess.
    // optional 'target' = '_blank','_self','_parent','_top' will add target attribute to the link.
    // For 'mode' => 'inline_node', optional 'inline_node_elements' defines allowable html tags to insert into, e.g.
    //    'inline_node_elements' => array('address', 'blockquote', 'cite', 'div', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p', 'span'),
    //    defaults to 'p'.
    //
    // Token replacement is performed on the following options using token types:
    //   'node'/'comment' - 'href', 'prefix', 'suffix', 'text', 'text_prefix', 'text_suffix', 'title', 'class', 'dropdown_name', 'dropdown_class'
    //   'link'           - 'href', 'prefix', 'suffix', 'text', 'text_prefix', 'text_suffix', 'title', 'class', 'dropdown_name', 'dropdown_class'
    // 'link' tokens are: [link-key], [link-text], [link-href], [link-language], [link-class], [link-rel], [link-title]
    // XSS filtering is performed on the following options: 'text' (if 'text_html' is set), 'text_prefix', 'text_suffix', 'dropdown_name'.

// Example:
//      // Use pattern '#itl_insert' to create new links
//      array('pattern' => '#itl_insert',     'mode' => 'link', 'text' => 'yahoo', 'class' => 'itweak-links-cssbutton', 'href' => 'http://yahoo.com', 'copy' => TRUE),
//      array('pattern' => '#itl_insert',     'mode' => 'link', 'text' => 'google', 'class' => 'itweak-links-cssbutton', 'href' => 'http://google.com', 'copy' => TRUE),
//      array('pattern' => '#itl_insert',     'mode' => 'link', 'text' => 'bing', 'class' => 'itweak-links-cssbutton', 'href' => 'http://bing.com'),

      // #BEGIN# Book links
      array('pattern' => 'outline_add_child', 'mode' => 'cssbutton'),
      // #END# Book links

      // #BEGIN# Forum Primary & Secondary links
      array('pattern' => 'forum',           'mode' => 'cssbutton', 'class' => 'itweak-links-cssbutton'),
      array('pattern' => 'mark-read',       'mode' => 'cssbutton', 'class' => 'itweak-links-cssbutton', 'text' => t('Mark all read')),
      array('pattern' => 'new-posts',       'mode' => 'link', ),
      array('pattern' => 'my-posts',        'mode' => 'link', ),
      array('pattern' => 'unanswered',      'mode' => 'link', 'text' => t('Unanswered'), ),
      array('pattern' => 'active',          'mode' => 'link', ),
      array('pattern' => 'login',           'mode' => 'link', ),
      // #END# Forum Primary & Secondary links

      // #BEGIN# Forum node/comment links
      array('pattern' => 'statistics_counter', 'mode' => 'link', ),

      // Links by various modules

      array('pattern' => 'flag-bookmarks',  'mode' => 'link', ),
      array('pattern' => 'fasttoggle_%',    'mode' => 'dropdown',
        'dropdown_class' => 'itweak-links-dropdown',
        'dropdown_name' => t('manage'),
        'class' => '', // link_class
      ),

      // "Subscribe" links by notifications.module
//      array('pattern' => 'notifications/subscribe/%/thread/nid/', 'key' => 'href', 'mode' => 'none', 'type' => 'nodecomment'), // Remove all notification links for replies to the comment. FIXME: Need 'Subscribe to: comments' notification?
//      array('pattern' => 'notifications/subscribe/%/nodetype/',   'key' => 'href', 'mode' => 'none'), // Remove all notification links for the node type
//      array('pattern' => 'notifications/subscribe/%',             'key' => 'href', 'mode' => 'link', ),
      array('pattern' => 'notifications/%subscribe/%',            'key' => 'href', 'mode' => 'dropdown',
        'dropdown_class' => 'itweak-links-dropdown',
        'dropdown_name' => t('subscribe'),
        'class' => '', // link_class
//        'class' => 'separator-before', // example style for dropdown menu separator
      ),

      // mollom.module
      array('pattern' => 'mollom_report',   'mode' => 'link', 'text' => t('spam')),

      // AF & Core Forum links
      array('pattern' => 'reply-allowed',   'mode' => 'cssbutton', 'class' => 'itweak-links-cssbutton', ),
      array('pattern' => 'comment_add',     'mode' => 'cssbutton', 'class' => 'itweak-links-cssbutton', 'text' => t('reply'), ),
      array('pattern' => 'comment_reply',   'mode' => 'cssbutton', 'class' => 'itweak-links-cssbutton', 'text' => t('reply'), ),
      array('pattern' => 'quote',           'mode' => 'cssbutton', 'class' => 'itweak-links-cssbutton', 'text' => t('quote'), ),
      array('pattern' => 'comment_edit',    'mode' => 'cssbutton', 'class' => 'itweak-links-cssbutton', 'text' => t('edit'),  ),
      array('pattern' => 'post_edit',       'mode' => 'cssbutton', 'class' => 'itweak-links-cssbutton', 'text' => t('edit'),  ),
      array('pattern' => 'comment_delete',  'mode' => 'cssbutton', 'class' => 'itweak-links-cssbutton', 'text' => t('delete'), ),
      array('pattern' => 'post_delete',     'mode' => 'cssbutton', 'class' => 'itweak-links-cssbutton', 'text' => t('delete'), ),

      // #END# Forum node/comment links

      // #BEGIN# Read More Link (ed_readmore.module for D6 and read_more.module for D7)
      array('pattern' => 'node_read_more',  'mode' => 'inline_node', 'class' => 'itweak-links-readmore',
        'type' => 'node',
        'prefix' => t('...'),
        'text' => t('<strong>Read more &raquo;</strong>'),
        'text_html' => TRUE,
        'title' => t('[title]'),
        'inline_node_elements' => array('address', 'blockquote', 'cite', 'div', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p', 'span'),
//        'copy' => TRUE, // Example how to replicate links. Note the next matching pattern
      ),
      array('pattern' => 'node_read_more',  'mode' => 'link', 'class' => 'itweak-links-readmore',
        'type' => 'node',
        'prefix' => t('...'),
        'text' => t('<strong>Read more &raquo;</strong>'),
        'text_html' => TRUE,
        'title' => t('[title]'),
      ),
//define('READ_MORE_TEXT_DEFAULT', '<strong>Read more<span class="element-invisible"> about [node:title]</span></strong>');
//define('READ_MORE_TITLE_DEFAULT', '[node:title]');
      // #END#


      // Leave other links as is, but move ahead of all processed
      array('pattern' => '%',               'mode' => 'link', 'weight' => -20),
    );
    $order = 0;
    $return = array();
    foreach ($settings as $key => $options) {
      $options['pattern'] = preg_replace('/[%]/', '.*', $options['pattern']);
      $options['pattern'] = preg_replace('/!/', '\!', $options['pattern']);
      $options += array('order' => $order, 'mode' => 'link', 'copy' => FALSE, 'nofollow' => TRUE);
      $return[] = $options;
      $order++;
    }
//drupal_set_message('DEBUG: preprocess_links settings='.iprintr($return));
  }
  return $return;
}

/*
 * Find all settings records that matches given link.
 */
function _itweak_links_preprocess_links_match($settings, $link_key, $link, $type=NULL, $node_type=NULL) {
  $result = array();

  // Scan through for a pattern match
  foreach ($settings as $key => $options) {
    if ($type && isset($options['type']) && $options['type'] != $type
      || $node_type && isset($options['node_type']) && $options['node_type'] != $node_type
    ) {
      continue;
    }
    $needle = $link_key;
    if (isset($options['key']) && $options['key'] != 'key') {
      if (!isset($link[$options['key']])) {
        continue;
      }
      $needle = $link[$options['key']];
    }
    if ($options['pattern'] == $needle || preg_match('!^' . $options['pattern'] . '!', $needle)) {
      $result[] = $options;
      if (empty($options['copy'])) {
        break;
      }
    }
  }
//drupal_set_message('DEBUG: match link_key='.$link_key.' result='.iprintr($result));
  return $result;
}

/**
 * @param $links
 *  Array of links to preprocess
 * @param $obj
 *  Comment or Node object (depends on $type).
 * @param $type
 *  Type of node to process the links for:
 *  'comment'     - standard comment links ($obj contains comment object)
 *  'node'        - standard node links ($obj contains node object)
 *  'nodecomment' - standard node links on a nodecomment ($obj contains node object)
 *  'primary'     - advanced forum primary links ($obj is NULL)
 *  'secondary'   - advanced forum secondary links ($obj is NULL)
 * @param $node_type
 *  Node content type (if applicable)
 */
function _itweak_links_preprocess_links(&$links, $obj, $type=NULL, $node_type=NULL) {
  if (!is_array($links)) return;
  if (isset($links['_itl_processed'])) return $links;    // Avoid double processing

  $settings = _itweak_links_preprocess_links_settings();

//drupal_set_message('DEBUG: preprocess_links type='.$type.' node_type='.$node_type.' links='.iprintr($links));
  $new_links = array();  // Create new array of links so we can change/append class names (keys)
  $weight = 0;
  $dropdowns = array();
  foreach ($links as $link_key => $link) {
    $orig_link = $link;
    $idx = '';
    $i = 0;

    $matches = _itweak_links_preprocess_links_match($settings, $link_key, $link, $type, $node_type);
    if (!empty($matches)) {
      foreach ($matches as $key => $options) {
        $link = $orig_link;
        $class = $link_key;
        $weight++;
        if ($options['mode'] == 'none') {
          // Remove the link
          continue;  // foreach
        }
        $attributes = isset($links['attributes']) ? $links['attributes'] : array();

        // Process tokens in select options
        if (!empty($obj) && module_exists('token') && function_exists('token_replace')) {
          $obj_type = ($type == 'comment') ? 'comment' : 'node';
          $link['key'] = $link_key;
          foreach (array(
            // token type => array(object, <field names>)
            $obj_type => array($obj,  'href', 'prefix', 'suffix', 'text', 'text_prefix', 'text_suffix', 'title', 'class', 'dropdown_name', 'dropdown_class'),
            'link'    => array($link, 'href', 'prefix', 'suffix', 'text', 'text_prefix', 'text_suffix', 'title', 'class', 'dropdown_name', 'dropdown_class'),
          ) as $t => $names) {
            $o = (object)array_shift($names);
            foreach ($names as $name) {
              if (isset($options[$name])) {
                $options[$name] = token_replace($options[$name], $t, $o);
              }
            }
          }
          unset($link['key']);
        }

        // Filter XSS in select options
        foreach (array('text' => 1, 'text_prefix' => 0, 'text_suffix' => 0, 'dropdown_name' => 0) as $name => $has_html_field) {
          if (isset($options[$name]) && (!$has_html_field || !empty($options[$name . '_html']) ) ) {
            $options[$name] = filter_xss_admin($options[$name]);
          }
        }

        // Verify element for link insertion
        if ($link_key == '#itl_insert') {
          $class = 'itl_link';
          if (!isset($options['href']) || !isset($options['text'])) {
            // Ignore malformed element
            continue;
          }
          if (empty($options['class'])) {
            $options['class'] = 'itl_link';
            $options['class_replace'] = TRUE;
          }
        }

        // Special class handling for 'mode' => 'cssbutton'
        $link_class_suffix = '';
        if ($options['mode'] == 'cssbutton') {
          $link_class_suffix = '-a';
          if (empty($options['class'])) {
            $options['class'] = 'itweak-links-cssbutton';
          }
        }

        // Override href of the link
        if (!empty($options['href'])) {
          $link['href'] = $options['href'];
        }

        // Apply options to the link
        if (!empty($options['text'])) {
          // Override title (it is a link's text actually, title attribute in <a> is for tooltips)
          $link['title'] = $options['text'];
          if (!empty($options['text_html'])) {
            $link['html'] = TRUE;
          }
          // else drupal will check_plain() $link['title']
        }
        if (!empty($options['title'])) {
          $attributes['title'] = $options['title'];
        }
        if (!empty($options['nofollow'])) {
          $attributes['rel'] = 'nofollow';
        }
        if (!empty($options['target'])) {
          $attributes['target'] = $options['target'];
        }
/* Not supported on D6
        if (!empty($options['prefix'])) {
          $link['#prefix'] = $options['prefix'];
        }
        if (!empty($options['suffix'])) {
          $link['#suffix'] = $options['suffix'];
        }
*/
//        $link_class = '';
        if (!empty($options['class'])) {
//          $link_class = $options['class'];
          if (empty($options['class_replace'])) {
            $class .= ' ' . $options['class'];
          }
          else {
            $class = $options['class'];
          }
          if (!empty($link['attributes']['class']) && empty($options['class_replace'])) {
            $attributes['class'] = $link['attributes']['class'] . ' ' . $options['class'] . $link_class_suffix;
          }
          else {
            $attributes['class'] = $options['class'] . $link_class_suffix;
          }
        }
        if (!empty($options['text_prefix'])) {
          $link['title'] = $options['text_prefix'] . $link['title'];
          $link['html'] = TRUE;
        }
        if (!empty($options['text_suffix'])) {
          $link['title'] = $link['title'] . $options['text_suffix'];
          $link['html'] = TRUE;
        }

//        if ($options['mode'] == 'cssbutton') {
//          $span_class = ($link_class != '') ? $link_class : 'itweak-links-cssbutton';
//          // Add extra span tags for image replacement.
//          $link['title'] .= '<span class="' . $span_class . '"></span>';
//          $link['html'] = TRUE;
//        }

        // Apply attributes
        if (!empty($attributes)) {
          if (empty($link['href'])) {
            if (!empty($link['html'])) {
              $link['title'] = _itweak_links_insert_attributes($link['title'], $attributes, 'a');
            }
          }
          else {
            $link['attributes'] = $attributes;
          }
        }

        // Process options specific to dropdown
        if ($options['mode'] == 'dropdown') {
          $dropdown_class = isset($options['dropdown_class']) ? $options['dropdown_class'] : 'itweak-links-dropdown';
          $dropdown_name = isset($options['dropdown_name']) ? $options['dropdown_name'] : 'dropdown';

          if (!isset($dropdowns[$dropdown_name])) {
            $dropdowns[$dropdown_name] = array(
              'weight' => $options['order'],  // Weight among other links - just use the link's order since it is first declaration of the dropdown
              'order'  => $options['order'],  // also copy order for second-pass sorting
                // FIXME: i'm somewhat fuzzy on the weight/order here... I feel like one is not needed here
              '#weight' => 0,  // saved weight for link ordering within this dropdown (equivalent to $weight)
              '#class' => $dropdown_class,
            );
          }
          else {
            $dropdowns[$dropdown_name]['#weight']++;
          }
        }

        // Adjust link weight
        if (!empty($options['weight'])) {
          // Override weight
          $link['weight'] = $options['weight'];
        } else if (!empty($options['order'])) {
          // Save order value for second-pass reordering
          $link['order'] = $options['order'];
          if ($options['mode'] == 'dropdown') {
            // Ordering will be within the dropdown
            $link += array('weight' => $dropdowns[$dropdown_name]['#weight']);
            $dropdowns[$dropdown_name]['#weight'] = $link['weight'];
          }
          else {
            $link += array('weight' => $weight);
            $weight = $link['weight'];
          }
        }

        // Move link to a dropdown
        if ($options['mode'] == 'dropdown') {
          $dropdowns[$dropdown_name][$class] = $link;
          continue; // foreach
        }

        // For links moving to node, save instructions for later - we will finish in node #pre_render
        if ($options['mode'] == 'before_node' || $options['mode'] == 'after_node' || $options['mode'] == 'inline_node') {
          $link['itl_prerender'] = $options;
        }
        $new_links[$class . $idx] = $link;
        $i++;
        $idx = ' itl_copy_' . $i;
      } // foreach ($matches ...)
    } else {
      // Default handling for links not found in the settings
      $link += array('weight' => $weight);
      $weight = $link['weight'];
      $new_links[$class . $idx] = $link;
    }
  } // foreach ($links ...)

  // Remove pseudo-element if by any chance it slipped through
  unset($new_links['#itl_insert']);

  // Now add the dropdowns in
  $classes_added = array(); // track what classes added to $new_links (avoids overwriting)
  foreach ($dropdowns as $dropdown_name => $dropdown) {
    $dropdown_weight = $dropdown['weight'];
    $dropdown_order = $dropdown['weight'];
    $dropdown_class = $dropdown['#class'];
    unset($dropdown['weight'], $dropdown['#weight'], $dropdown['#class'], $dropdown['order']);
    // Sort links within the dropdown
    uasort($dropdown, '_itweak_links_link_cmp');
    // This generates ul list:
    $html = theme('links', $dropdown, array('class' => $dropdown_class . '-menu'));
    // Build drop-down markup structure here:
    $html ='<span class="' . $dropdown_class . '-title-wrap">'
      . '<a href="#" rel="nofollow" onclick="javascript: return false;">' . $dropdown_name . '</a>'
      . '</span>' . $html;
    // FIXME: Add dropdown title to the above
    // FIXME: Create a theme hook?

    $i = 0;
    $check = $dropdown_class;
    while (in_array($check, $classes_added)) {
      $check = $dropdown_class . ' i' . $i++; // Add increment class
    }
    $dropdown_class = $check;
    $new_links[$dropdown_class] = array(
      'weight' => $dropdown_weight,
      'order' => $dropdown_order,
      'html' => TRUE,
      'title' => $html,
      'attributes' => array('class' => $dropdown_class . '-title'),
    );
    $classes_added[] = $dropdown_class;
  }

  // $weight value is preserved
  $weight++;
  foreach ($new_links as $key => $link) {
    if (isset($link['order'])) {
      $new_links[$key]['weight'] = $weight + $link['order'];
      unset($new_links[$key]['order']);
    }
    if (!isset($min_weight) || $min_weight > $new_links[$key]['weight']) {
      $min_weight = $new_links[$key]['weight'];
    }
  }
  if (!isset($min_weight)) {
    $min_weight = 0;
  }

  // Create a phantom element to mark the list as already processed
  $new_links['_itl_processed'] = array(
    'weight' => $min_weight + 0.001,
  );
  uasort($new_links, '_itweak_links_link_cmp');
/*
drupal_set_message('DEBUG: OUT preprocess_links(' . $type .')'
.'<table><tr>'
.'<tr><th>Input<th>Output</tr>'
.'<td>Total '.count($links).':'.iprintr($links)
.'<td>Total '.count($new_links).':'.iprintr($new_links)
.'</tr></table>'
); */
  $links = $new_links;
}

//END